package com.sillycycle.bagleyd.loop;

/*
 * @(#)Loop.java
 *
 * Copyright 2010 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Cursor;
import java.awt.MenuItem;
import java.awt.Point;
import java.awt.PopupMenu;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.swing.JFrame;

/**
 * The <code>Loop</code> class extends the
 * <code>LoopCanvas</code> for handling the responses to the user
 * for the <code>LoopFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

class Loop extends LoopCanvas implements Runnable,
		MouseListener, MouseMotionListener, KeyListener {
	private static final long serialVersionUID = 42L;

	boolean paused = true;
	int now = 0;
	byte cellState = 0;
	boolean quit = false;
	private transient Thread gameRunner;
	PopupMenu popupMenu = null;
	transient LoopPopup popupListener = null;

	public Loop(JFrame frame, int speed, int columns, int neighbors,
			boolean randomOrient, boolean grid,
			boolean wrap, boolean visible, Color bg) {
		this.frame = frame;
		this.speed = speed;
		this.neighbors = neighbors;
		this.ncols = getRealColumns(columns);
		this.nrows = getRowsFromColumns(columns);
		this.randomOrient = randomOrient;
		this.grid = grid;
		this.wrap = wrap;
		this.visible = visible;
		this.background = bg;
		setBackground(bg);
		this.addKeyListener(this);
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		setCursor(new Cursor(Cursor.HAND_CURSOR));
		initPopup();
		start();
	}

	public void setPopupMenu(PopupMenu popupMenu) {
		this.popupMenu = popupMenu;
		add(popupMenu);
	}

	@Override
	protected void processMouseEvent(MouseEvent event) {
		if ((popupMenu != null) && event.isPopupTrigger())
			popupMenu.show(this, event.getX(), event.getY());
		else
			super.processMouseEvent(event);
	}

	protected final void initPopup() {
		popupListener = new LoopPopup(this);
		popupMenu = new PopupMenu("Select a template");

		MenuItem menuItem = new MenuItem("Cut");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Copy");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Paste");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Clear");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		popupMenu.addSeparator();

		menuItem = new MenuItem("Cw");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Ccw");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Refl Wall");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Refl Grnd");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		setPopupMenu(popupMenu);
	}

	void quitGame() {
		banner(LoopInterface.MSG_END);
		quit = true;
	}

	synchronized boolean processAdvance() {
		boolean gotNew = false;

		if (!paused) {
			loopPulse(visible, speed == LoopInterface.MAX_SPEED);
		}
		return gotNew;
	}

	boolean processKey(char key) {
		char charPressed = key;

		switch (charPressed) {
		case '>':
		case '.':
			if (speed < LoopInterface.MAX_SPEED) {
				speedUp();
			}
			return true;
		case '<':
		case ',':
			if (speed > LoopInterface.MIN_SPEED) {
				slowDown();
			}
			return true;
		case '?':
			((LoopFrame) frame).descriptionHelp();
			return true;
		case '!':
			((LoopFrame) frame).featuresHelp();
			return true;
		case '^':
			((LoopFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((LoopFrame) frame).aboutHelp();
			return true;
		case 'g':
			getLoop();
			break;
		case 'w':
			writeLoop();
			break;
		case 'G':
			getLoopFile();
			break;
		case 'W':
			writeLoopFile();
			break;
		case 'p':
		case 'P':
			printLoop();
			break;
		case 'c':
		case 'C':
			setPause(true);
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
				//e.printStackTrace();
			}
			clearField(true);
			break;
		case ']':
		case '}':
			ncols++;
			((LoopFrame) frame).setColumns(ncols);
			break;
		case '[':
		case '{':
			if (ncols > 4) {
				ncols--;
				((LoopFrame) frame).setColumns(ncols);
			}
			break;
		case 'y':
		case 'Y':
			if (paused) {
				capturer.clearCapture();
				loopPulse(visible, false);
			} else {
				setPause(true);
				((LoopFrame) frame).updateGen(gen);
				((LoopFrame) frame).updateCells(nCells);
			}
			break;
		case '#':
			grid = !grid;
			drawGrid();
			break;
		case 'z':
		case 'Z':
			randomOrient = !randomOrient;
			break;
		case '|':
		case '\\':
		case 'r':
		case 'R':
			capturer.clearCapture();
			((LoopFrame) frame).updateGen(gen);
			((LoopFrame) frame).updateCells(nCells);
			setPause(!paused);
			banner("");
			break;
		case 'o':
		case 'O':
			wrap = !wrap;
			break;
		case 'v':
		case 'V':
			visible = !visible;
			break;
		case 's':
		case 'S':
			((LoopFrame) frame).pauseGame();
			break;
		case '&':
			((LoopFrame) frame).clearGame();
			break;
		case '~':
			((LoopFrame) frame).setupGame();
			break;
		case '*':
			((LoopFrame) frame).showBuffer();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		//case '\03': /* Control-C */
			quitGame();
			break;
		default:
			return false;
		}
		return true;
	}

	void redrawAll() {
		repaint();
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (mouseDown)
			return;

		Point p = cellPl.screenToCell(event.getX(), event.getY());

		if (p == null)
			return;
		mouseDown = true;
		if ((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) {
			if ((modifiers & InputEvent.BUTTON1_DOWN_MASK) != 0) {
				capturer.clearCapture();
				if (p.x >= 0 && p.x < ncols &&
						p.y >= 0 && p.y < nrows) {
					cellState = incCell(p.x, p.y);
				}
			}
		} else {
			capturer.clearCapture();
			capturer.startCapture(p.x, p.y);
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		if (mouseDown) { // Do not care which mouse button
			mouseDown = false;
		}
		capturer.stopCapture();
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		focus = true;
		drawFrame(focus);
		//banner(null);
	}

	@Override
	public void mouseExited(MouseEvent event) {
		focus = false;
		drawFrame(focus);
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if ((modifiers & InputEvent.BUTTON3_DOWN_MASK) == 0) {
			Point p = cellPl.screenToCell(event.getX(), event.getY());

			if (p == null)
				return;
			mouseDown = true;
			if ((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) {
				capturer.clearCapture();
				if (p.x >= 0 && p.x < ncols &&
						p.y >= 0 && p.y < nrows) {
					setCell(p.x, p.y, cellState, true);
				}
			} else {
				capturer.continueCapture(p.x, p.y);
			}
		}
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	public final void start() {
		if (gameRunner == null) {
			gameRunner = new Thread(this);
			gameRunner.start();
		}
	}

	public void stop() {
		/*if (gameRunner.isAlive())
			gameRunner.stop();*/
		gameRunner = null;
	}

	@Override
	public void run() {
		inits();
		/*sizeInit();*/
		repaint();
		while (!quit) {
			try {
				processAdvance();
				if (Thread.interrupted()) {
					throw new InterruptedException();
				}
				int delay = 20 * (LoopInterface.MAX_SPEED +
					LoopInterface.MIN_SPEED - speed) - 19;

				if (delay > 0)
					Thread.sleep(delay);
			} catch (InterruptedException e) {
				//e.printStackTrace();
			}
		}
		((LoopFrame) frame).quit();
	}

	@Override
	public void keyPressed(KeyEvent event) {
		if (event.getKeyCode() == KeyEvent.VK_ESCAPE) {
			((LoopFrame) frame).shuffleDown();
			return;
		}
		processKey(event.getKeyChar());
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	public boolean getPause() {
		return paused;
	}

	public void setPause(boolean paused) {
		if (paused == this.paused)
			return;
		this.paused = paused;
		if (paused)
			banner(LoopInterface.MSG_PAUSED);
		else
			banner("");
	}

	public synchronized void setDelay(int delay) {
		this.speed = LoopInterface.MAX_SPEED +
			LoopInterface.MIN_SPEED - 1 - delay;
	}

	public void setNeighborIndex(int neighborIndex) {
		neighbors = LoopRules.neighborToIndex(neighborIndex);
		if (grid)
			drawGrid();
	}

	public void setRandomOrient(boolean flag) {
		this.randomOrient = flag;
	}

	public void setGrid(boolean flag) {
		this.grid = flag;
		drawGrid();
	}

	public void setWrap(boolean flag) {
		this.wrap = flag;
	}

	public void setCellVisible(boolean flag) {
		this.visible = flag;
	}

	public void setColumns(int columns) {
		if (this.ncols != getRealColumns(columns) ||
			this.nrows != getRowsFromColumns(columns)) {
			eraseFrame();
			this.ncols = getRealColumns(columns);
			this.nrows = getRowsFromColumns(columns);
			/*sizeInit();*/
			resizeField();
			drawField(true);
		}
	}
}
